# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np

from . import group_accuracy_score, group_balanced_root_mean_squared_error
from . import group_fallout_rate, group_max_error
from . import group_mean_absolute_error, group_mean_overprediction
from . import group_mean_squared_error, group_mean_squared_log_error
from . import group_mean_underprediction, group_median_absolute_error
from . import group_mean_prediction
from . import group_miss_rate, group_precision_score, group_r2_score
from . import group_recall_score, group_roc_auc_score, group_root_mean_squared_error
from . import group_selection_rate, group_specificity_score, group_zero_one_loss

_GROUP_NAMES_MSG = "The group_names property must be a list of strings"
_METRICS_KEYS_MSG = "Keys for metrics dictionary must be strings"
_METRICS_VALUES_MSG = "Values for metrics dictionary must be of type GroupMetricResult"

_ARRAYS_NOT_SAME_LENGTH = "Lengths of y_true, y_pred and groups must match"
_BIN_MISMATCH_FOR_METRIC = "The groups for metric {0} do not match the groups property"
_GROUP_NAMES_BAD_COUNT = "Count of group_names not the same as the number of unique groups"

_Y_TRUE = 'trueY'
_Y_PRED = 'predictedY'
_PRECOMPUTED_METRICS = 'precomputedMetrics'
_GLOBAL = 'global'
_BINS = 'bins'
_PRECOMPUTED_BINS = 'precomputedFeatureBins'
_BIN_VECTOR = 'binVector'
_BIN_LABELS = 'binLabels'
_FEATURE_BIN_NAME = 'featureBinName'
_PREDICTION_TYPE = 'predictionType'
_PREDICTION_BINARY_CLASSIFICATION = 'binaryClassification'
_MODEL_NAMES = 'modelNames'
_SCHEMA = 'schemaType'
_GROUP_METRIC_SET = 'groupMetricSet'
_VERSION = 'schemaVersion'

_UNSUPPORTED_MODEL_TYPE = "The specified model_type of '{0}' is not supported"
_DICT_TOO_MANY_Y_PRED = 'Too many y_pred values in dictionary'

BINARY_CLASSIFICATION = 'binary_classification'
REGRESSION = 'regression'
_allowed_model_types = frozenset([BINARY_CLASSIFICATION, REGRESSION])

# The following keys need to match those of _metric_methods in
# _fairlearn_dashboard.py
# Issue 269 is about unifying the two sets
GROUP_ACCURACY_SCORE = "accuracy_score"
GROUP_BALANCED_ROOT_MEAN_SQUARED_ERROR = "balanced_root_mean_squared_error"
GROUP_FALLOUT_RATE = "fallout_rate"
GROUP_MAX_ERROR = "max_error"
GROUP_MEAN_ABSOLUTE_ERROR = "mean_absolute_error"
GROUP_MEAN_OVERPREDICTION = "overprediction"
GROUP_MEAN_PREDICTION = "average"
GROUP_MEAN_SQUARED_ERROR = "mean_squared_error"
GROUP_MEAN_SQUARED_LOG_ERROR = "6d106114-4433-40a2-b091-8983ab540a53"
GROUP_MEAN_UNDERPREDICTION = "underprediction"
GROUP_MEDIAN_ABSOLUTE_ERROR = "median_absolute_error"
GROUP_MISS_RATE = "miss_rate"
GROUP_PRECISION_SCORE = "precision_score"
GROUP_R2_SCORE = "r2_score"
GROUP_RECALL_SCORE = "recall_score"
GROUP_ROC_AUC_SCORE = "balanced_accuracy_score"
GROUP_ROOT_MEAN_SQUARED_ERROR = "root_mean_squared_error"
GROUP_SELECTION_RATE = "selection_rate"
GROUP_SPECIFICITY_SCORE = "specificity_score"
GROUP_ZERO_ONE_LOSS = "zero_one_loss"

BINARY_CLASSIFICATION_METRICS = {}
BINARY_CLASSIFICATION_METRICS[GROUP_ACCURACY_SCORE] = group_accuracy_score
BINARY_CLASSIFICATION_METRICS[GROUP_FALLOUT_RATE] = group_fallout_rate
BINARY_CLASSIFICATION_METRICS[GROUP_MEAN_OVERPREDICTION] = group_mean_overprediction
BINARY_CLASSIFICATION_METRICS[GROUP_MEAN_UNDERPREDICTION] = group_mean_underprediction
BINARY_CLASSIFICATION_METRICS[GROUP_MISS_RATE] = group_miss_rate
BINARY_CLASSIFICATION_METRICS[GROUP_PRECISION_SCORE] = group_precision_score
BINARY_CLASSIFICATION_METRICS[GROUP_RECALL_SCORE] = group_recall_score
BINARY_CLASSIFICATION_METRICS[GROUP_ROC_AUC_SCORE] = group_roc_auc_score
BINARY_CLASSIFICATION_METRICS[GROUP_SELECTION_RATE] = group_selection_rate
BINARY_CLASSIFICATION_METRICS[GROUP_SPECIFICITY_SCORE] = group_specificity_score

REGRESSION_METRICS = {}
REGRESSION_METRICS[GROUP_BALANCED_ROOT_MEAN_SQUARED_ERROR] = group_balanced_root_mean_squared_error  # noqa:E501
REGRESSION_METRICS[GROUP_MAX_ERROR] = group_max_error
REGRESSION_METRICS[GROUP_MEAN_ABSOLUTE_ERROR] = group_mean_absolute_error
REGRESSION_METRICS[GROUP_MEAN_OVERPREDICTION] = group_mean_overprediction
REGRESSION_METRICS[GROUP_MEAN_PREDICTION] = group_mean_prediction
REGRESSION_METRICS[GROUP_MEAN_SQUARED_ERROR] = group_mean_squared_error
REGRESSION_METRICS[GROUP_MEAN_SQUARED_LOG_ERROR] = group_mean_squared_log_error
REGRESSION_METRICS[GROUP_MEAN_UNDERPREDICTION] = group_mean_underprediction
REGRESSION_METRICS[GROUP_MEDIAN_ABSOLUTE_ERROR] = group_median_absolute_error
REGRESSION_METRICS[GROUP_R2_SCORE] = group_r2_score
REGRESSION_METRICS[GROUP_ROOT_MEAN_SQUARED_ERROR] = group_root_mean_squared_error
REGRESSION_METRICS[GROUP_ZERO_ONE_LOSS] = group_zero_one_loss


def create_group_metric_set(model_type,
                            y_true,
                            y_preds,
                            group_memberships,
                            model_titles=None,
                            group_titles=None,
                            extra_metrics=None):
    """Create a dictionary matching the Dashboard's cache."""
    if extra_metrics is not None:
        raise NotImplementedError("No support for extra_metrics yet")

    # We could consider checking that the length of y_preds matches model_titles
    # and that the length of group_memberships matches group_titles

    result = dict()
    result[_SCHEMA] = _GROUP_METRIC_SET
    result[_VERSION] = 0

    if model_type not in _allowed_model_types:
        msg_format = "model_type '{0}' not in {1}"
        msg = msg_format.format(model_type, sorted(
            list(_allowed_model_types)))
        raise ValueError(msg)

    function_dict = None
    if model_type == BINARY_CLASSIFICATION:
        result[_PREDICTION_TYPE] = _PREDICTION_BINARY_CLASSIFICATION
        function_dict = BINARY_CLASSIFICATION_METRICS
    else:
        raise NotImplementedError("No support yet for regression")

    _yt = np.asarray(y_true)
    result[_Y_TRUE] = _yt.tolist()

    result[_Y_PRED] = []
    result[_PRECOMPUTED_METRICS] = []
    result[_PRECOMPUTED_BINS] = []
    result[_MODEL_NAMES] = []
    for g, group_membership in enumerate(group_memberships):
        _gm = np.asarray(group_membership).tolist()
        _unique_groups = sorted(list(np.unique(_gm)))
        group_names = [str(x) for x in _unique_groups]
        groups = [_unique_groups.index(x) for x in _gm]
        bin_dict = {_BIN_VECTOR: groups, _BIN_LABELS: group_names}
        if group_titles is not None:
            bin_dict[_FEATURE_BIN_NAME] = group_titles[g]
        result[_PRECOMPUTED_BINS].append(bin_dict)

        model_list = []
        for m, model_pred in enumerate(y_preds):
            _yp = np.asarray(model_pred).tolist()

            # Only record each y_pred and model name once
            if g == 0:
                result[_Y_PRED].append(_yp)
                if model_titles is not None:
                    result[_MODEL_NAMES].append(model_titles[m])

            metric_dict = dict()
            for metric_key, metric_func in function_dict.items():
                gmr = metric_func(_yt, _yp, groups)
                curr_dict = dict()
                curr_dict[_GLOBAL] = gmr.overall
                curr_dict[_BINS] = list(gmr.by_group.values())
                metric_dict[metric_key] = curr_dict
            model_list.append(metric_dict)
        result[_PRECOMPUTED_METRICS].append(model_list)

    return result



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""This module contains algorithms implementing the reductions approach to disparity mitigation.

In this approach, disparity constraints are cast as Lagrange multipliers, which cause the
reweighting and relabelling of the input data. This *reduces* the problem back to standard machine
learning training.
"""

from ._reduction import Reduction  # noqa: F401

from ._exponentiated_gradient import ExponentiatedGradient  # noqa: F401
from ._exponentiated_gradient import ExponentiatedGradientResult  # noqa: F401
from ._grid_search import GridSearch, GridSearchResult  # noqa: F401
from ._moments import AbsoluteLoss, Moment, ConditionalSelectionRate  # noqa: F401
from ._moments import DemographicParity, EqualizedOdds, ErrorRate   # noqa: F401
from ._moments import GroupLossMoment, SquareLoss, ZeroOneLoss  # noqa: F401
from ._moments import ClassificationMoment, LossMoment  # noqa: F401

_exponentiated_gradient = [
    "ExponentiatedGradient",
    "ExponentiatedGradientResult"
]

_grid_search = [
    "GridSearch",
    "GridSearchResult"
]

_moments = [
    "AbsoluteLoss",
    "Moment",
    "ClassificationMoment",
    "ConditionalSelectionRate",
    "DemographicParity",
    "EqualizedOdds",
    "ErrorRate",
    "GroupLossMoment",
    "LossMoment",
    "SquareLoss",
    "ZeroOneLoss"
]

__all__ = ["Reduction"] + _exponentiated_gradient + _grid_search + _moments



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Package for Exponentiated Gradient."""

from .exponentiated_gradient import ExponentiatedGradient  # noqa: F401
from ._exponentiated_gradient_result import ExponentiatedGradientResult  # noqa: F401

__all__ = [
    "ExponentiatedGradient",
    "ExponentiatedGradientResult"
]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging
import numpy as np
import pandas as pd
from fairlearn.reductions import Reduction
from ._constants import _ACCURACY_MUL, _REGRET_CHECK_START_T, _REGRET_CHECK_INCREASE_T, \
    _SHRINK_REGRET, _SHRINK_ETA, _MIN_T, _RUN_LP_STEP, _PRECISION, _INDENTATION
from ._lagrangian import _Lagrangian
from ._exponentiated_gradient_result import ExponentiatedGradientResult
from fairlearn._input_validation import _validate_and_reformat_input

logger = logging.getLogger(__name__)


def _mean_pred(X, hs, weights):
    """Return a weighted average of predictions produced by classifiers in `hs`."""
    pred = pd.DataFrame()
    for t in range(len(hs)):
        pred[t] = hs[t](X)
    return pred[weights.index].dot(weights)


class ExponentiatedGradient(Reduction):
    """An Estimator which implements the exponentiated gradient approach to reductions.

    The exponentiated gradient algorithm is described in detail by
    `Agarwal et al. (2018) <https://arxiv.org/abs/1803.02453>`_.

    :param estimator: An estimator implementing methods :code:`fit(X, y, sample_weight)` and
        :code:`predict(X)`, where `X` is the matrix of features, `y` is the vector of labels, and
        `sample_weight` is a vector of weights; labels `y` and predictions returned by
        :code:`predict(X)` are either 0 or 1.
    :type estimator: estimator

    :param constraints: The disparity constraints expressed as moments
    :type constraints: fairlearn.reductions.Moment

    :param eps: Allowed fairness constraint violation; the solution best_classifier is
        guaranteed to have the classification error within :code:`2*best_gap` of the best error
        under constraint eps; the constraint violation is at most :code:`2*(eps+best_gap)`
    :type eps: float

    :param T: Maximum number of iterations
    :type T: int

    :param nu: Convergence threshold for the duality gap, corresponding to a
        conservative automatic setting based on the statistical uncertainty in measuring
        classification error
    :type nu: float

    :param eta_mul: Initial setting of the learning rate
    :type eta_mul: float
    """

    def __init__(self, estimator, constraints, eps=0.01, T=50, nu=None, eta_mul=2.0):  # noqa: D103
        self._estimator = estimator
        self._constraints = constraints
        self._eps = eps
        self._T = T
        self._nu = nu
        self._eta_mul = eta_mul
        self._best_classifier = None
        self._classifiers = None

    def fit(self, X, y, **kwargs):
        """Return a fair classifier under specified fairness constraints.

        :param X: The feature matrix
        :type X: numpy.ndarray or pandas.DataFrame

        :param y: The label vector
        :type y: numpy.ndarray, pandas.DataFrame, pandas.Series, or list
        """
        X_train, y_train, A = _validate_and_reformat_input(X, y, **kwargs)

        n = X_train.shape[0]

        logger.debug("...Exponentiated Gradient STARTING")

        B = 1 / self._eps
        lagrangian = _Lagrangian(X_train, A, y_train, self._estimator, self._constraints,
                                 self._eps, B)

        theta = pd.Series(0, lagrangian.constraints.index)
        Qsum = pd.Series()
        lambdas = pd.DataFrame()
        gaps_EG = []
        gaps = []
        Qs = []

        last_regret_checked = _REGRET_CHECK_START_T
        last_gap = np.PINF
        for t in range(0, self._T):
            logger.debug("...iter=%03d", t)

            # set lambdas for every constraint
            lambda_vec = B * np.exp(theta) / (1 + np.exp(theta).sum())
            lambdas[t] = lambda_vec
            lambda_EG = lambdas.mean(axis=1)

            # select classifier according to best_h method
            h, h_idx = lagrangian.best_h(lambda_vec)
            pred_h = h(X_train)

            if t == 0:
                if self._nu is None:
                    self._nu = _ACCURACY_MUL * (pred_h - y_train).abs().std() / np.sqrt(n)
                eta_min = self._nu / (2 * B)
                eta = self._eta_mul / B
                logger.debug("...eps=%.3f, B=%.1f, nu=%.6f, T=%d, eta_min=%.6f",
                             self._eps, B, self._nu, self._T, eta_min)

            if h_idx not in Qsum.index:
                Qsum.at[h_idx] = 0.0
            Qsum[h_idx] += 1.0
            gamma = lagrangian.gammas[h_idx]
            Q_EG = Qsum / Qsum.sum()
            result_EG = lagrangian.eval_gap(Q_EG, lambda_EG, self._nu)
            gap_EG = result_EG.gap()
            gaps_EG.append(gap_EG)

            if t == 0 or not _RUN_LP_STEP:
                gap_LP = np.PINF
            else:
                # saddle point optimization over the convex hull of
                # classifiers returned so far
                Q_LP, _, result_LP = lagrangian.solve_linprog(self._nu)
                gap_LP = result_LP.gap()

            # keep values from exponentiated gradient or linear programming
            if gap_EG < gap_LP:
                Qs.append(Q_EG)
                gaps.append(gap_EG)
            else:
                Qs.append(Q_LP)
                gaps.append(gap_LP)

            logger.debug("%seta=%.6f, L_low=%.3f, L=%.3f, L_high=%.3f"
                         ", gap=%.6f, disp=%.3f, err=%.3f, gap_LP=%.6f",
                         _INDENTATION, eta, result_EG.L_low,
                         result_EG.L, result_EG.L_high,
                         gap_EG, result_EG.gamma.max(),
                         result_EG.error, gap_LP)

            if (gaps[t] < self._nu) and (t >= _MIN_T):
                # solution found
                break

            # update regret
            if t >= last_regret_checked * _REGRET_CHECK_INCREASE_T:
                best_gap = min(gaps_EG)

                if best_gap > last_gap * _SHRINK_REGRET:
                    eta *= _SHRINK_ETA
                last_regret_checked = t
                last_gap = best_gap

            # update theta based on learning rate
            theta += eta * (gamma - self._eps)

        self._expgrad_result = self._format_results(gaps, Qs, lagrangian, B, eta_min)

        self._best_classifier = self._expgrad_result._best_classifier
        self._classifiers = self._expgrad_result._classifiers

    def predict(self, X):
        """Provide a prediction for the given input data.

        Note that this is non-deterministic, due to the nature of the
        exponentiated gradient algorithm.

        :param X: Feature data
        :type X: numpy.ndarray or pandas.DataFrame

        :return: The prediction. If `X` represents the data for a single example
            the result will be a scalar. Otherwise the result will be a vector
        :rtype: Scalar or vector
        """
        positive_probs = self._best_classifier(X)
        return (positive_probs >= np.random.rand(len(positive_probs))) * 1

    def _pmf_predict(self, X):
        """Probability mass function for the given input data.

        :param X: The data for which predictions are required
        :type X: Array
        :return: Array of tuples with the probabilities of predicting 0 and 1.
        :rtype: Array
        """
        positive_probs = self._best_classifier(X)
        return np.concatenate((1-positive_probs, positive_probs), axis=1)

    def _format_results(self, gaps, Qs, lagrangian, B, eta_min):
        gaps_series = pd.Series(gaps)
        gaps_best = gaps_series[gaps_series <= gaps_series.min() + _PRECISION]
        best_t = gaps_best.index[-1]
        weights = Qs[best_t]
        hs = lagrangian.hs
        for h_idx in hs.index:
            if h_idx not in weights.index:
                weights.at[h_idx] = 0.0

        def best_classifier(X): return _mean_pred(X, hs, weights)
        best_gap = gaps[best_t]

        last_t = len(Qs) - 1

        result = ExponentiatedGradientResult(
            best_classifier,
            best_gap,
            lagrangian,
            weights,
            last_t,
            best_t)

        logger.debug("...eps=%.3f, B=%.1f, nu=%.6f, T=%d, eta_min=%.6f",
                     self._eps, B, self._nu, self._T, eta_min)
        logger.debug("...last_t=%d, best_t=%d, best_gap=%.6f, n_oracle_calls=%d, n_hs=%d",
                     last_t, best_t, best_gap, lagrangian.n_oracle_calls,
                     len(lagrangian.classifiers))

        return result



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""This module contains methods which operate on a predictor, rather than an estimator.

The predictor's output is adjusted to fulfill specified parity constraints. The postprocessors
learn how to adjust the predictor's output from the training data.
"""

from ._threshold_optimizer import ThresholdOptimizer  # noqa: F401

__all__ = [
    "ThresholdOptimizer"
]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import math
import pytest

import fairlearn.metrics as metrics


def test_all_correct():
    y_true = [0, 0, 0, 0, 1, 1, 1, 1]
    y_pred = [0, 0, 0, 0, 1, 1, 1, 1]

    result = metrics.balanced_root_mean_squared_error(y_true, y_pred)

    assert result == 0


def test_all_incorrect():
    y_true = [0, 0, 0, 0, 1, 1, 1, 1]
    y_pred = [1, 1, 1, 1, 0, 0, 0, 0]

    result = metrics.balanced_root_mean_squared_error(y_true, y_pred)

    assert result == 1


def test_weighted():
    y_true = [0, 0, 0, 0, 1, 1, 1, 1]
    y_pred = [0, 0, 0, 1, 0, 0, 0, 1]
    weight = [1, 2, 3, 4, 5, 6, 7, 8]

    result = metrics.balanced_root_mean_squared_error(y_true, y_pred, weight)

    expected_0 = math.sqrt(4.0/10)
    expected_1 = math.sqrt(18.0/26)

    assert (expected_0+expected_1)/2 == result


def test_probability_predict():
    y_true = [0, 0, 0, 0, 1]
    y_pred = [0.5, 0.5, 0.5, 0.5, 0.75]

    result = metrics.balanced_root_mean_squared_error(y_true, y_pred)

    expected_0 = 0.5
    expected_1 = 0.25

    assert (expected_0+expected_1)/2 == result


@pytest.mark.parametrize("y_true", [[0, 0, 0], [1, 1, 1], [0, 0, 2], [1, 1, 2], [0, 1, 2]])
def test_y_true_not_0_1(y_true):
    y_pred = [0.5, 0.5, 0.5]

    message = "Only 0 and 1 are allowed in y_true and both must be present"
    with pytest.raises(ValueError) as execInfo:
        _ = metrics.balanced_root_mean_squared_error(y_true, y_pred)
    assert message == execInfo.value.args[0]
